from typing import TypeVar

from pydantic import BaseModel

from ...types.connection import TileConnectionType
from ...types.tile import TileType

KT = TypeVar("KT")
VT = TypeVar("VT")


class AbstractTile[KT, VT](BaseModel):
    id: str
    name: str | None = None
    rbac: dict = {}
    tile_type: TileType = TileType.UNKNOWN
    next: dict[TileConnectionType, str] = {}  # str=id
    contents: dict[KT, VT] = {}
    ids: set[KT] = set()



import requests

from ....sdk.contracts.dtos.sdk.requests.world.get import WorldGetRequest
from ....sdk.contracts.dtos.sdk.responses.response import Response
from ....sdk.contracts.dtos.sdk.responses.world.get import WorldGetResponse
from ..abstract import AbstractCommand


class WorldGetCommand(AbstractCommand):
    async def execute(self, request: WorldGetRequest) -> WorldGetResponse:
        response: requests.Response = requests.get(
            url=f"http://{self.options.tld}/world/{request.id}", timeout=self.options.timeout
        )
        return Response[WorldGetResponse].model_validate(response.json()).data



import time

from pydantic import BaseModel

from ..types.entity import EntityType


class Entity(BaseModel):
    id: str
    entity_type: EntityType
    updated: float | None = None  # unix timestamp
    rbac: dict = {}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # set creation time as last time updated
        self.updated = time.time()



import logging
import os
import shutil
import uuid
from pathlib import Path

from pydantic import BaseModel

from ...sdk.contracts.dtos.island import Island
from ...sdk.contracts.dtos.sdk.wrapped_data import WrappedData
from ...sdk.contracts.errors.server.dao.conflict import DaoConflictError
from ...sdk.contracts.errors.server.dao.doesnotexist import DaoDoesNotExistError
from ...sdk.contracts.errors.server.dao.inconsistency import DaoInconsistencyError

logger = logging.getLogger()


class IslandDao(BaseModel):
    # ["base"]  / "worlds" / "world_id" / "islands" / "island_id" / metadata.json
    storage_base_path: Path

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.storage_base_path.mkdir(parents=True, exist_ok=True)

    ### Internal ##################################

    def _island_metadata_path(self, world_id: str, island_id: str) -> Path:
        # ["base"] / "worlds" / "world_id" / "islands" / "island_id" / metadata.json
        return self.storage_base_path / "worlds" / world_id / "islands" / island_id / "metadata.json"

    async def get(self, world_id: str, island_id: str) -> WrappedData[Island]:
        logger.debug("[IslandDAO] getting island metadata from storage")
        island_metadata_path: Path = self._island_metadata_path(world_id=world_id, island_id=island_id)
        if not island_metadata_path.exists():
            raise DaoDoesNotExistError("island metadata does not exist")
        with open(file=island_metadata_path, mode="r", encoding="utf-8") as file:
            os.fsync(file)
            json_data: str = file.read()
        return WrappedData[Island].model_validate_json(json_data)

    async def post(self, world_id: str, island: Island) -> None:
        logger.debug("[IslandDAO] posting island metadata to storage")
        island_metadata_path: Path = self._island_metadata_path(world_id=world_id, island_id=island.id)
        if island_metadata_path.exists():
            raise DaoConflictError("island metadata already exists")
        if not island_metadata_path.parents[2].exists():
            raise DaoDoesNotExistError("island container (world) does not exist")
        if not island_metadata_path.parent.exists():
            logger.debug("[IslandDAO] island metadata folder creating ..")
            island_metadata_path.parent.mkdir(parents=True, exist_ok=True)

        nonce: str = str(uuid.uuid4())
        wrapped_data: WrappedData[Island] = WrappedData[Island](data=island, nonce=nonce)
        wrapped_data_raw: str = wrapped_data.model_dump_json(exclude={"data": {"next", "contents"}})
        with open(file=island_metadata_path, mode="w", encoding="utf-8") as file:
            file.write(wrapped_data_raw)
            os.fsync(file)

        # now validate we stored
        stored_island: WrappedData[Island] = await self.get(world_id=world_id, island_id=island.id)
        if stored_island.nonce != nonce:
            msg: str = f"storage inconsistency detected while storing island {island.id} - nonce mismatch!"
            raise DaoInconsistencyError(msg)

    async def put_safe(self, world_id: str, wrapped_island: WrappedData[Island]) -> None:
        logger.debug("[IslandDAO] putting island data to storage")
        island_metadata_path: Path = self._island_metadata_path(world_id=world_id, island_id=wrapped_island.data.id)
        if not island_metadata_path.parents[2].exists():
            raise DaoDoesNotExistError("island container (world) does not exist")
        if not island_metadata_path.parent.exists():
            logger.debug("[IslandDAO] island metadata folder creating ..")
            island_metadata_path.parent.mkdir(parents=True, exist_ok=True)

        # see if we have a pre-existing nonce to verify against
        try:
            previous_state = await self.get(world_id=world_id, island_id=wrapped_island.data.id)
            if previous_state.nonce != wrapped_island.nonce:
                msg: str = (
                    f"storage inconsistency detected while putting island {wrapped_island.data.id} - nonce mismatch!"
                )
                raise DaoInconsistencyError(msg)
        except DaoDoesNotExistError:
            # then no nonce to verify against.
            pass

        # if we made it this far we are safe to update

        nonce: str = str(uuid.uuid4())
        wrapped_data: WrappedData[Island] = WrappedData[Island](data=wrapped_island.data, nonce=nonce)
        wrapped_data_raw: str = wrapped_data.model_dump_json(exclude={"data": {"next", "contents"}})
        with open(file=island_metadata_path, mode="w", encoding="utf-8") as file:
            file.write(wrapped_data_raw)
            os.fsync(file)

        # now validate we stored
        stored_island: WrappedData[Island] = await self.get(world_id=world_id, island_id=wrapped_data.data.id)
        if stored_island.nonce != nonce:
            msg: str = (
                f"storage inconsistency detected while verifying put island {wrapped_data.data.id} - nonce mismatch!"
            )
            raise DaoInconsistencyError(msg)

    async def delete(self, world_id: str, island_id: str) -> None:
        logger.debug("[IslandDAO] deleting island data from storage")
        island_metadata_path: Path = self._island_metadata_path(world_id=world_id, island_id=island_id)
        if not island_metadata_path.exists():
            raise DaoDoesNotExistError("island metadata does not exist")
        # remove "island_id"/ and lower
        shutil.rmtree(island_metadata_path.parent)



from ..types.tile import TileType
from .island_full import IslandFull
from .tiles.abtract import AbstractTile


class WorldFull(AbstractTile[str, IslandFull]):
    """Basic World"""

    tile_type: TileType = TileType.WORLD



from ..types.tile import TileType
from .tiles.abtract import AbstractTile
from .tiles.tile import Tile


class IslandFull(AbstractTile[str, Tile]):
    tile_type: TileType = TileType.ISLAND
    dimensions: tuple[int, int] | None = None
    biome: TileType | None = None



# Excluding server from package exports (explicitly)

from .client.client import Client
from .client.commands.abstract import AbstractCommand
from .client.commands.island.create import IslandCreateCommand
from .client.commands.island.delete import IslandDeleteCommand
from .client.commands.island.get import IslandGetCommand
from .client.commands.metrics.health.get import HealthGetCommand
from .client.commands.world.create import WorldCreateCommand
from .client.commands.world.delete import WorldDeleteCommand
from .client.commands.world.get import WorldGetCommand
from .sdk.common.config.environment import (
    demand_env_var,
    demand_env_var_as_bool,
    demand_env_var_as_float,
    demand_env_var_as_int,
    get_env_var,
)
from .sdk.contracts.dtos.coordinate import Coordinate
from .sdk.contracts.dtos.entity import Entity
from .sdk.contracts.dtos.island import Island
from .sdk.contracts.dtos.island_full import IslandFull
from .sdk.contracts.dtos.sdk.command_options import CommandOptions
from .sdk.contracts.dtos.sdk.request_status_codes import RequestStatusCodes
from .sdk.contracts.dtos.sdk.requests.island.create import IslandCreateRequest
from .sdk.contracts.dtos.sdk.requests.island.delete import IslandDeleteRequest
from .sdk.contracts.dtos.sdk.requests.island.get import IslandGetRequest
from .sdk.contracts.dtos.sdk.requests.world.create import WorldCreateRequest
from .sdk.contracts.dtos.sdk.requests.world.delete import WorldDeleteRequest
from .sdk.contracts.dtos.sdk.requests.world.get import WorldGetRequest
from .sdk.contracts.dtos.sdk.responses.island.create import IslandCreateResponse
from .sdk.contracts.dtos.sdk.responses.island.delete import IslandDeleteResponse
from .sdk.contracts.dtos.sdk.responses.island.get import IslandGetResponse
from .sdk.contracts.dtos.sdk.responses.response import Response
from .sdk.contracts.dtos.sdk.responses.world.create import WorldCreateResponse
from .sdk.contracts.dtos.sdk.responses.world.get import WorldGetResponse
from .sdk.contracts.dtos.sdk.wrapped_data import WrappedData
from .sdk.contracts.dtos.sdk.wrapped_request import WrappedRequest
from .sdk.contracts.dtos.tiles.abtract import AbstractTile
from .sdk.contracts.dtos.tiles.tile import Tile
from .sdk.contracts.dtos.window import Window
from .sdk.contracts.dtos.world import World
from .sdk.contracts.dtos.world_full import WorldFull
from .sdk.contracts.errors.common.environment_variable_not_found import EnvironmentVariableNotFoundError
from .sdk.contracts.errors.sdk.exceeded_retry_count import ExceededRetryCountError
from .sdk.contracts.errors.sdk.request_failure import RequestFailureError
from .sdk.contracts.errors.sdk.unknown_verb import UnknownVerbError
from .sdk.contracts.errors.server.dao.conflict import DaoConflictError
from .sdk.contracts.errors.server.dao.doesnotexist import DaoDoesNotExistError
from .sdk.contracts.errors.server.dao.inconsistency import DaoInconsistencyError
from .sdk.contracts.errors.server.factory import FactoryError
from .sdk.contracts.errors.server.service import ServiceError
from .sdk.contracts.types.connection import TileConnectionType
from .sdk.contracts.types.entity import EntityType
from .sdk.contracts.types.sdk.request_verb import RequestVerbType
from .sdk.contracts.types.tile import TileType



import uuid

from pydantic import BaseModel

from ....sdk.contracts.dtos.world import World
from ...dao.world import WorldDao


class WorldFactory(BaseModel):
    worlddao: WorldDao

    async def create(self, name: str | None = None) -> str:
        if name is None:
            name = "darkness"
        world: World = World(id=str(uuid.uuid4()), name=name)
        await self.worlddao.post(world=world)
        return world.id



import logging
import os
import shutil
import uuid
from pathlib import Path

from pydantic import BaseModel

from ...sdk.contracts.dtos.sdk.wrapped_data import WrappedData
from ...sdk.contracts.dtos.world import World
from ...sdk.contracts.errors.server.dao.conflict import DaoConflictError
from ...sdk.contracts.errors.server.dao.doesnotexist import DaoDoesNotExistError
from ...sdk.contracts.errors.server.dao.inconsistency import DaoInconsistencyError

logger = logging.getLogger()


class WorldDao(BaseModel):
    # ["base"]  / "worlds" / "world_id" / metadata.json
    storage_base_path: Path

    def _world_metadata_path(self, world_id: str) -> Path:
        # ["base"] / "worlds" / "world_id" / metadata.json
        return self.storage_base_path / "worlds" / world_id / "metadata.json"

    ### Internal ##################################

    async def get(self, world_id: str) -> WrappedData[World]:
        logger.debug("[WorldDAO] getting world metadata from storage")
        world_metadata_path: Path = self._world_metadata_path(world_id=world_id)
        if not world_metadata_path.exists():
            raise DaoDoesNotExistError("world metadata does not exist")
        with open(file=world_metadata_path, mode="r", encoding="utf-8") as file:
            os.fsync(file)
            json_data: str = file.read()
        return WrappedData[World].model_validate_json(json_data)

    async def post(self, world: World) -> None:
        logger.debug("[WorldDAO] posting world metadata to storage")
        world_metadata_path: Path = self._world_metadata_path(world_id=world.id)
        if world_metadata_path.exists():
            raise DaoConflictError("world metadata already exists")
        if not world_metadata_path.parent.exists():
            logger.debug("[WorldDAO] world metadata folder creating ..")
            world_metadata_path.parent.mkdir(parents=True, exist_ok=True)

        nonce: str = str(uuid.uuid4())
        wrapped_data: WrappedData[World] = WrappedData[World](data=world, nonce=nonce)
        wrapped_data_raw: str = wrapped_data.model_dump_json(exclude={"data": {"next", "contents"}})
        with open(file=world_metadata_path, mode="w", encoding="utf-8") as file:
            file.write(wrapped_data_raw)
            os.fsync(file)

        # now validate we stored
        stored_world: WrappedData[World] = await self.get(world_id=world.id)
        if stored_world.nonce != nonce:
            msg: str = f"storage inconsistency detected while storing world {world.id} - nonce mismatch!"
            raise DaoInconsistencyError(msg)

    async def put_safe(self, wrapped_world: WrappedData[World]) -> None:
        world_id: str = wrapped_world.data.id

        logger.debug("[WorldDAO] putting world data to storage")
        world_metadata_path: Path = self._world_metadata_path(world_id=world_id)
        if not world_metadata_path.parent.exists():
            logger.debug("[WorldDAO] world metadata folder creating ..")
            world_metadata_path.parent.mkdir(parents=True, exist_ok=True)

        # see if we have a pre-existing nonce to verify against
        try:
            previous_state: WrappedData[World] = await self.get(world_id=world_id)
            if previous_state.nonce != wrapped_world.nonce:
                msg: str = f"storage inconsistency detected while putting world {world_id} - nonce mismatch!"
                raise DaoInconsistencyError(msg)
        except DaoDoesNotExistError:
            # then no nonce to verify against.
            pass

        # if we made it this far we are safe to update

        nonce: str = str(uuid.uuid4())
        wrapped_data: WrappedData[World] = WrappedData[World](data=wrapped_world.data, nonce=nonce)
        wrapped_data_raw: str = wrapped_data.model_dump_json(exclude={"data": {"next", "contents"}})
        with open(file=world_metadata_path, mode="w", encoding="utf-8") as file:
            file.write(wrapped_data_raw)
            os.fsync(file)

        # now validate we stored
        stored_world: WrappedData[World] = await self.get(world_id=world_id)
        if stored_world.nonce != nonce:
            msg: str = (
                f"storage inconsistency detected while verifying put world {wrapped_data.data.id} - nonce mismatch!"
            )
            raise DaoInconsistencyError(msg)

    async def delete(self, world_id: str) -> None:
        logger.debug("[WorldDAO] deleting world data from storage")
        world_metadata_path: Path = self._world_metadata_path(world_id=world_id)
        if not world_metadata_path.exists():
            raise DaoDoesNotExistError("world metadata does not exist")
        # remove "world_id"/ and lower
        shutil.rmtree(world_metadata_path.parent)



from abc import abstractmethod

from pydantic import BaseModel

from ....sdk.contracts.types.tile import TileType
from ...dao.island import IslandDao


class AbstractIslandFactory(BaseModel):
    islanddao: IslandDao

    @abstractmethod
    async def create(self, world_id: str, name: str | None, dimensions: tuple[int, int], biome: TileType) -> str:
        pass



from enum import Enum


class TileType(str, Enum):
    WORLD = "world"
    ISLAND = "island"

    UNKNOWN = "unknown"

    OCEAN = "ocean"
    WATER = "water"

    SHORE = "shore"
    DIRT = "dirt"
    ROCK = "rock"
    GRASS = "grass"
    FOREST = "forest"



import requests

from ....sdk.contracts.dtos.sdk.requests.world.create import WorldCreateRequest
from ....sdk.contracts.dtos.sdk.responses.response import Response
from ....sdk.contracts.dtos.sdk.responses.world.create import WorldCreateResponse
from ..abstract import AbstractCommand


class WorldCreateCommand(AbstractCommand):
    async def execute(self, request: WorldCreateRequest) -> WorldCreateResponse:
        response: requests.Response = requests.post(
            url=f"http://{self.options.tld}/world",
            timeout=self.options.timeout,
            json=request.model_dump(),
        )
        return Response[WorldCreateResponse].model_validate(response.json()).data



import requests

from ....sdk.contracts.dtos.sdk.requests.island.delete import IslandDeleteRequest
from ....sdk.contracts.dtos.sdk.responses.island.delete import IslandDeleteResponse
from ....sdk.contracts.dtos.sdk.responses.response import Response
from ..abstract import AbstractCommand


class IslandDeleteCommand(AbstractCommand):
    async def execute(self, request: IslandDeleteRequest) -> IslandDeleteResponse:
        response: requests.Response = requests.delete(
            url=f"http://{self.options.tld}/world/{request.world_id}/island/{request.island_id}",
            timeout=self.options.timeout,
        )
        return Response[IslandDeleteResponse].model_validate(response.json()).data



import requests

from ....sdk.contracts.dtos.sdk.requests.island.create import IslandCreateRequest
from ....sdk.contracts.dtos.sdk.responses.island.create import IslandCreateResponse
from ....sdk.contracts.dtos.sdk.responses.response import Response
from ..abstract import AbstractCommand


class IslandCreateCommand(AbstractCommand):
    async def execute(self, request: IslandCreateRequest) -> IslandCreateResponse:
        response: requests.Response = requests.post(
            url=f"http://{self.options.tld}/world/{request.world_id}/island",
            timeout=self.options.timeout,
            json=request.model_dump(),
        )
        return Response[IslandCreateResponse].model_validate(response.json()).data



import requests

from ....sdk.contracts.dtos.sdk.requests.island.get import IslandGetRequest
from ....sdk.contracts.dtos.sdk.responses.island.get import IslandGetResponse
from ....sdk.contracts.dtos.sdk.responses.response import Response
from ..abstract import AbstractCommand


class IslandGetCommand(AbstractCommand):
    async def execute(self, request: IslandGetRequest) -> IslandGetResponse:
        response: requests.Response = requests.get(
            url=f"http://{self.options.tld}/world/{request.world_id}/island/{request.island_id}",
            timeout=self.options.timeout,
            params={"full": request.full},
        )
        return Response[IslandGetResponse].model_validate(response.json()).data



from ..types.tile import TileType
from .island import Island
from .tiles.abtract import AbstractTile


class World(AbstractTile[str, Island]):
    """Basic World"""

    tile_type: TileType = TileType.WORLD



import asyncio
import logging
import re
import secrets
import uuid
from asyncio import Queue

from ....sdk.contracts.dtos.coordinate import Coordinate
from ....sdk.contracts.dtos.island import Island
from ....sdk.contracts.dtos.sdk.wrapped_data import WrappedData
from ....sdk.contracts.dtos.tiles.tile import Tile
from ....sdk.contracts.dtos.window import Window
from ....sdk.contracts.errors.server.dao.inconsistency import DaoInconsistencyError
from ....sdk.contracts.types.connection import TileConnectionType
from ....sdk.contracts.types.tile import TileType
from ...dao.tile import TileDao
from .abstract import AbstractIslandFactory

logger = logging.getLogger()


class FlatIslandFactory(AbstractIslandFactory):
    tiledao: TileDao

    @staticmethod
    async def producer(window: Window, queue: Queue):
        range_x_min: int = window.min.x - 1
        range_x_max: int = window.max.x
        range_y_min: int = window.min.x - 1
        range_y_max: int = window.max.y

        for x in range(range_x_min, range_x_max):
            for y in range(range_y_min, range_y_max):
                local_x = x + 1
                local_y = y + 1
                tile_id: str = f"tile_{local_x}_{local_y}"
                await queue.put(tile_id)

    async def tiles_process(self, world_id: str, island: Island, window: Window) -> None:
        # Lets shovel in some biome default or dirt tiles!
        async def step_one():
            async def consumer(queue: Queue):
                while not queue.empty():
                    local_tile_id: str = await queue.get()
                    # Mutate tiles to biome default (or dirt)
                    await self.mutate_tile(
                        world_id=world_id,
                        island_id=island.id,
                        tile_id=local_tile_id,
                        mutate=90,  # percentage of 100%
                        tile_type=(island.biome if island.biome else TileType.DIRT),
                    )
                    queue.task_done()

            queue = asyncio.Queue()
            await asyncio.gather(FlatIslandFactory.producer(window, queue), consumer(queue))

        await step_one()

        # Add Eradicates (rocks)
        async def step_two():
            async def consumer(queue: Queue):
                while not queue.empty():
                    local_tile_id: str = await queue.get()
                    await self.mutate_tile(
                        world_id=world_id,
                        island_id=island.id,
                        tile_id=local_tile_id,
                        mutate=0.5,  # 0.5% change (very low)
                        tile_type=TileType.ROCK,
                    )
                    queue.task_done()

            queue = asyncio.Queue()
            await asyncio.gather(FlatIslandFactory.producer(window, queue), consumer(queue))

        await step_two()

        # TODO: isolated ocean is NOT ocean, we MUST have path to the edge
        # Convert inner Ocean to Water Tiles
        async def step_three():
            async def consumer(queue: Queue):
                while not queue.empty():
                    local_tile_id: str = await queue.get()
                    # Convert inner Ocean to Water Tiles
                    await self.brackish_tile(world_id=world_id, island_id=island.id, tile_id=local_tile_id)
                    queue.task_done()

            queue = asyncio.Queue()
            await asyncio.gather(FlatIslandFactory.producer(window, queue), consumer(queue))

        await step_three()

        # Erode Tiles (to make shore)
        async def step_four():
            async def consumer(queue: Queue):
                while not queue.empty():
                    local_tile_id: str = await queue.get()
                    await self.erode_tile(world_id=world_id, island_id=island.id, tile_id=local_tile_id)
                    queue.task_done()

            queue = asyncio.Queue()
            await asyncio.gather(FlatIslandFactory.producer(window, queue), consumer(queue))

        await step_four()

        # Grow Tiles
        async def step_five():
            async def consumer(queue: Queue):
                while not queue.empty():
                    local_tile_id: str = await queue.get()
                    await self.grow_tile(world_id=world_id, island_id=island.id, tile_id=local_tile_id)
                    queue.task_done()

            queue = asyncio.Queue()
            await asyncio.gather(FlatIslandFactory.producer(window, queue), consumer(queue))

        await step_five()

    async def mutate_tile(
        self, world_id: str, island_id: str, tile_id: str, mutate: float, tile_type: TileType
    ) -> None:
        if secrets.randbelow(100) <= mutate:
            # then we spawn the tile type

            # We don't current have a patch, so get and put..
            # get
            target_tile: WrappedData[Tile] = await self.tiledao.get(
                world_id=world_id, island_id=island_id, tile_id=tile_id
            )
            # patch
            target_tile.data.tile_type = tile_type
            # put
            await self.tiledao.put_safe(world_id=world_id, island_id=island_id, wrapped_tile=target_tile)

    async def convert_tile(
        self, world_id: str, island_id: str, tile_id: str, source: TileType, target: TileType
    ) -> None:
        # get
        target_tile: WrappedData[Tile] = await self.tiledao.get(world_id=world_id, island_id=island_id, tile_id=tile_id)

        # check
        if target_tile.data.tile_type == source:
            # patch
            target_tile.data.tile_type = target

            # put
            await self.tiledao.put_safe(world_id=world_id, island_id=island_id, wrapped_tile=target_tile)

    async def adjecent_liquids(self, world_id: str, island_id: str, tile_id: str) -> list[TileType]:
        return await self.adjecent_to(
            world_id=world_id, island_id=island_id, tile_id=tile_id, types=[TileType.OCEAN, TileType.WATER]
        )

    async def adjecent_to(
        self, world_id: str, island_id: str, tile_id: str, types: list[TileType] | None
    ) -> list[TileType]:
        adjecent_targets: list[TileType] = []

        target_tile: WrappedData[Tile] = await self.tiledao.get(world_id=world_id, island_id=island_id, tile_id=tile_id)
        for _, adjecent_id in target_tile.data.next.items():

            adjecent_tile: WrappedData[Tile] = await self.tiledao.get(
                world_id=world_id, island_id=island_id, tile_id=adjecent_id
            )
            # logger.debug(f"tile {tile_id} -> {adjecent_dir} -> {adjecent_id} -> {adjecent_tile.data.tile_type}")

            if adjecent_tile.data.tile_type in types and adjecent_tile.data.tile_type not in adjecent_targets:
                adjecent_targets.append(adjecent_tile.data.tile_type)
        # logger.debug(f"tile {tile_id} is adjecent to: {adjecent_targets}")
        return adjecent_targets

    # async def get_adjecent_tiles(
    #     self,
    #     world_id: str,
    #     island_id: str,
    #     tile_id: str,
    #     types: list[TileType] | None = None
    # ) -> list[WrappedData[Tile]]:
    #     adjecent_targets: list[WrappedData[Tile]] = []
    #
    #     target_tile: WrappedData[Tile] = await self.tiledao.get(world_id=world_id, island_id=island_id, tile_id=tile_id)
    #     for _, adjecent_id in target_tile.data.next.items():
    #
    #         adjecent_tile: WrappedData[Tile] = await self.tiledao.get(
    #             world_id=world_id, island_id=island_id, tile_id=adjecent_id
    #         )
    #         # logger.debug(f"tile {tile_id} -> {adjecent_dir} -> {adjecent_id} -> {adjecent_tile.data.tile_type}")
    #
    #         # if omitted then include all types
    #         if types is not None:
    #             types: list[TileType] = [TileType(local_type) for local_type in TileType]
    #
    #         if adjecent_tile.data.tile_type in types:
    #             adjecent_targets.append(adjecent_tile)
    #     # logger.debug(f"tile {tile_id} is adjecent to: {adjecent_targets}")
    #     return adjecent_targets

    async def grow_tile(self, world_id: str, island_id: str, tile_id: str) -> None:
        # if island.tiles[tile_id].tile_type not in [TileType.UNKNOWN, TileType.OCEAN, TileType.WATER]:

        # get
        target_tile: WrappedData[Tile] = await self.tiledao.get(world_id=world_id, island_id=island_id, tile_id=tile_id)

        # dirt -> grass
        if target_tile.data.tile_type == TileType.DIRT:
            adjecent_liquids: list[TileType] = await self.adjecent_liquids(
                world_id=world_id, island_id=island_id, tile_id=tile_id
            )
            if TileType.WATER in adjecent_liquids and TileType.OCEAN not in adjecent_liquids:
                # patch
                target_tile.data.tile_type = TileType.GRASS

                # put
                await self.tiledao.put_safe(world_id=world_id, island_id=island_id, wrapped_tile=target_tile)

                # get updated local tile
                target_tile = await self.tiledao.get(world_id=world_id, island_id=island_id, tile_id=tile_id)

        # grass+water (no dirt/ocean) -> forest
        if target_tile.data.tile_type == TileType.GRASS:
            neighbors: list[TileType] = await self.adjecent_to(
                world_id=world_id,
                island_id=island_id,
                tile_id=tile_id,
                types=[TileType.WATER, TileType.GRASS, TileType.OCEAN],
            )

            # no forests grow by oceans
            if TileType.OCEAN in neighbors:
                return

            if len(neighbors) > 1:
                # next to more than one kind (grass/water)

                # patch
                target_tile.data.tile_type = TileType.FOREST

                # put
                try:
                    await self.tiledao.put_safe(world_id=world_id, island_id=island_id, wrapped_tile=target_tile)
                except DaoInconsistencyError as error:
                    logger.error(target_tile.model_dump_json())
                    raise error

        # # grass+(dirt)
        # if island.tiles[tile_id].tile_type == TileType.GRASS:
        #     neighbors: list[TileType] = FlatIslandFactory.adjecent_to(
        #         # TODO: review this josh ...
        #         island=island, tile_id=tile_id, types=[TileType.WATER, TileType.GRASS, TileType.DIRT, TileType.FOREST]
        #     )

    async def brackish_tile(self, world_id: str, island_id: str, tile_id: str):
        # Convert inner Ocean to Water Tiles

        # See if we are next to another ocean tile
        neighbors: list[TileType] = await self.adjecent_to(
            world_id=world_id, island_id=island_id, tile_id=tile_id, types=[TileType.OCEAN]
        )
        if len(neighbors) < 1:
            await self.convert_tile(
                world_id=world_id, island_id=island_id, tile_id=tile_id, source=TileType.OCEAN, target=TileType.WATER
            )

        # # are we an isolated ocean body? if so then we are water
        #
        # # 1. For each next that is liquid (recusively), can we get to the edge?
        # root_tile_id: str = tile_id
        #
        # # async def check_if_ocean(root_tile_id: str) -> bool:
        # #     # 1. Are we ocean? - if no then return False
        # #     target_tile: WrappedData[Tile] = await self.tiledao.get(world_id=world_id, island_id=island_id, tile_id=root_tile_id)
        # #     if target_tile.data.tile_type != TileType.OCEAN:
        # #         return False
        # #
        # #     # 2. Are we edge? - if yes then return True
        # #     #
        # #     # 1d space
        # #     # nexts? len=0, then - True
        # #     #
        # #     # 2d space
        # #     # nexts? len=1->4,
        # #     # if nexts < max of 4, then - True
        # #     if len(target_tile.data.next) < 4:
        # #         return True
        # #
        # #     # local = [ [conn, con_id] for conn in target_tile.data.next]
        # #
        # #     # 3. then check my nexts
        # #     #       if check all nexts is True, then True
        # #
        # #     # 4. Return False (not ocean)
        # #     return False
        #
        # if not await check_if_ocean(root_tile_id=root_tile_id):
        #     # if True then we are water,
        #     await self.convert_tile(
        #         world_id=world_id, island_id=island_id, tile_id=tile_id, source=TileType.OCEAN, target=TileType.WATER
        #     )
        #
        #     # root_tile: Tile = (await self.tiledao.get(world_id=world_id, island_id=island_id, tile_id=root_tile_id)).data

    async def erode_tile(self, world_id: str, island_id: str, tile_id: str) -> None:
        msg: str = f"eroding tile: {tile_id}"
        logger.debug(msg)

        # get
        target_tile: WrappedData[Tile] = await self.tiledao.get(world_id=world_id, island_id=island_id, tile_id=tile_id)

        # shore erosion
        if target_tile.data.tile_type not in [TileType.UNKNOWN, TileType.OCEAN, TileType.WATER, TileType.SHORE]:
            adjecent_liquids: list[TileType] = await self.adjecent_liquids(
                world_id=world_id, island_id=island_id, tile_id=tile_id
            )
            # Apply erosion - rocks and be left by oceans, everything else becomes shore
            if TileType.OCEAN in adjecent_liquids:
                if target_tile.data.tile_type not in (TileType.ROCK, TileType.SHORE):
                    previous_type = target_tile.data.tile_type

                    # patch
                    target_tile.data.tile_type = TileType.SHORE

                    # put
                    await self.tiledao.put_safe(world_id=world_id, island_id=island_id, wrapped_tile=target_tile)

                    msg: str = f"tile {tile_id}:{previous_type} converted to {TileType.SHORE}"
                    logger.debug(msg)

    async def generate_ocean_block(self, world_id: str, island_id: str, window: Window):
        # 1. fill a blank nXm area with ocean
        async def step_one():
            async def consumer(queue: Queue):
                while not queue.empty():
                    local_tile_id: str = await queue.get()
                    local_tile: Tile = Tile(id=local_tile_id, tile_type=TileType.OCEAN)

                    # create tile
                    await self.tiledao.post(world_id=world_id, island_id=island_id, tile=local_tile)
                    msg: str = f"({local_tile_id}) brought into existence as {TileType.OCEAN}"
                    logger.debug(msg)

                    # Update the island --

                    # get
                    wrapped_island: WrappedData[Island] = await self.islanddao.get(
                        world_id=world_id, island_id=island_id
                    )

                    # Patch - Add tile_id to in-memory representation before storing
                    wrapped_island.data.ids.add(local_tile_id)

                    # put -- store island update (tile addition)
                    await self.islanddao.put_safe(world_id=world_id, wrapped_island=wrapped_island)

                    queue.task_done()

            queue = asyncio.Queue()
            await asyncio.gather(FlatIslandFactory.producer(window, queue), consumer(queue))

        await step_one()

        # 2. Connect everything together
        async def step_two():
            wrapped_island: WrappedData[Island] = await self.islanddao.get(world_id=world_id, island_id=island_id)
            pattern: re.Pattern = re.compile("^tile_([a-zA-Z0-9-]+)_([a-zA-Z0-9-]+)$")

            async def consumer(queue):
                while not queue.empty():
                    local_tile_id: str = await queue.get()
                    match = pattern.search(local_tile_id)
                    # tile_id: str = f"tile_{local_x}_{local_y}"
                    local_x: int = int(match.group(1))
                    local_y: int = int(match.group(2))

                    # msg: str = f"binding ({local_tile_id}) physically to peers"
                    # logger.debug(msg)

                    # Bind LEFT
                    target_tile_id: str = f"tile_{local_x - 1}_{local_y}"
                    if target_tile_id in wrapped_island.data.ids:
                        # load tile
                        local_tile: WrappedData[Tile] = await self.tiledao.get(
                            world_id=world_id, island_id=wrapped_island.data.id, tile_id=local_tile_id
                        )

                        # update tile
                        local_tile.data.next[TileConnectionType.LEFT] = target_tile_id
                        msg: str = (
                            f"tile {local_tile_id} -> {TileConnectionType.LEFT} -> {local_tile.data.next[TileConnectionType.LEFT]}"
                        )
                        logger.debug(msg)

                        # store tile
                        await self.tiledao.put_safe(
                            world_id=world_id, island_id=wrapped_island.data.id, wrapped_tile=local_tile
                        )

                    # Bind RIGHT
                    target_tile_id: str = f"tile_{local_x + 1}_{local_y}"
                    if target_tile_id in wrapped_island.data.ids:
                        # load tile
                        local_tile: WrappedData[Tile] = await self.tiledao.get(
                            world_id=world_id, island_id=wrapped_island.data.id, tile_id=local_tile_id
                        )

                        # update tile
                        local_tile.data.next[TileConnectionType.RIGHT] = target_tile_id

                        msg: str = (
                            f"tile {local_tile_id} -> {TileConnectionType.RIGHT} -> {local_tile.data.next[TileConnectionType.RIGHT]}"
                        )
                        logger.debug(msg)

                        # store tile
                        await self.tiledao.put_safe(
                            world_id=world_id, island_id=wrapped_island.data.id, wrapped_tile=local_tile
                        )

                    # Bind UP
                    target_tile_id: str = f"tile_{local_x}_{local_y - 1}"
                    if target_tile_id in wrapped_island.data.ids:
                        # load tile
                        local_tile: WrappedData[Tile] = await self.tiledao.get(
                            world_id=world_id, island_id=wrapped_island.data.id, tile_id=local_tile_id
                        )

                        # update tile
                        local_tile.data.next[TileConnectionType.UP] = target_tile_id
                        msg: str = (
                            f"tile {local_tile_id} -> {TileConnectionType.UP} -> {local_tile.data.next[TileConnectionType.UP]}"
                        )
                        logger.debug(msg)

                        # store tile
                        await self.tiledao.put_safe(
                            world_id=world_id, island_id=wrapped_island.data.id, wrapped_tile=local_tile
                        )

                    # Bind DOWN
                    target_tile_id: str = f"tile_{local_x}_{local_y + 1}"
                    if target_tile_id in wrapped_island.data.ids:
                        # load tile
                        local_tile: WrappedData[Tile] = await self.tiledao.get(
                            world_id=world_id, island_id=wrapped_island.data.id, tile_id=local_tile_id
                        )

                        # update tile
                        local_tile.data.next[TileConnectionType.DOWN] = target_tile_id
                        msg: str = (
                            f"tile {local_tile_id} -> {TileConnectionType.DOWN} -> {local_tile.data.next[TileConnectionType.DOWN]}"
                        )
                        logger.debug(msg)

                        # store tile
                        await self.tiledao.put_safe(
                            world_id=world_id, island_id=wrapped_island.data.id, wrapped_tile=local_tile
                        )

                    queue.task_done()

            queue = asyncio.Queue()
            await asyncio.gather(FlatIslandFactory.producer(window, queue), consumer(queue))

        await step_two()

    async def create(self, world_id: str, name: str | None, dimensions: tuple[int, int], biome: TileType) -> Island:
        if name is None:
            name = "roshar"

        # 1. blank, named island
        island: Island = Island(id=str(uuid.uuid4()), name=name, dimensions=dimensions, biome=biome)
        await self.islanddao.post(world_id=world_id, island=island)

        # Define the maximum size
        max_x, max_y = dimensions

        # Generate an empty 2D block of ocean
        window: Window = Window(min=Coordinate(x=1, y=1), max=Coordinate(x=max_x, y=max_y))
        await self.generate_ocean_block(world_id=world_id, island_id=island.id, window=window)

        # Apply our terrain generation
        window = Window(min=Coordinate(x=2, y=2), max=Coordinate(x=max_x - 1, y=max_y - 1))
        await self.tiles_process(world_id=world_id, island=island, window=window)

        # get final state and return
        return (await self.islanddao.get(world_id=world_id, island_id=island.id)).data



import logging
import sys
from pathlib import Path

from ...sdk.contracts.errors.server.service import ServiceError
from ..dao.island import IslandDao
from ..dao.tile import TileDao
from ..dao.world import WorldDao
from ..services.state import StateService

logger = logging.getLogger()

STORAGE_BASE_PATH: Path = Path(".") / "data"


class ContextManager:
    # Application Level Services
    state_service: StateService | None = None

    def __init__(self):
        if ContextManager.state_service is None:
            ContextManager.state_service = StateService(
                worlddao=WorldDao(storage_base_path=STORAGE_BASE_PATH),
                islanddao=IslandDao(storage_base_path=STORAGE_BASE_PATH),
                tiledao=TileDao(storage_base_path=STORAGE_BASE_PATH),
            )
            logger.debug("[ContextManager] assigned new state service to context manager")
        else:
            logger.debug("[ContextManager] state service already assigned")


try:
    ContextManager()
except ServiceError as error:
    logger.error(error)
    sys.exit(1)



import logging

from pydantic import BaseModel

from ...sdk.contracts.dtos.island import Island
from ...sdk.contracts.dtos.island_full import IslandFull
from ...sdk.contracts.dtos.sdk.requests.island.create import IslandCreateRequest
from ...sdk.contracts.dtos.sdk.requests.island.delete import IslandDeleteRequest
from ...sdk.contracts.dtos.sdk.requests.island.get import IslandGetRequest
from ...sdk.contracts.dtos.sdk.requests.world.create import WorldCreateRequest
from ...sdk.contracts.dtos.sdk.requests.world.delete import WorldDeleteRequest
from ...sdk.contracts.dtos.sdk.requests.world.get import WorldGetRequest
from ...sdk.contracts.dtos.sdk.wrapped_data import WrappedData
from ...sdk.contracts.dtos.tiles.tile import Tile
from ...sdk.contracts.dtos.world import World
from ...sdk.contracts.dtos.world_full import WorldFull
from ..dao.island import IslandDao
from ..dao.tile import TileDao
from ..dao.world import WorldDao
from ..factories.island.flat import FlatIslandFactory
from ..factories.world.world import WorldFactory

logger = logging.getLogger()


class StateService(BaseModel):
    worlddao: WorldDao
    islanddao: IslandDao
    tiledao: TileDao

    worldfactory: WorldFactory | None = None
    flatislandfactory: FlatIslandFactory | None = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.worldfactory = WorldFactory(worlddao=self.worlddao)
        self.flatislandfactory = FlatIslandFactory(islanddao=self.islanddao, tiledao=self.tiledao)

    ### World ##################################

    async def world_create(self, request: WorldCreateRequest) -> str:
        logger.debug("[StateService] creating world")
        return await self.worldfactory.create(name=request.name)

    async def world_lite_get(self, request: WorldGetRequest) -> World:
        return (await self.worlddao.get(world_id=request.id)).data

    async def world_get(self, request: WorldGetRequest) -> WorldFull:
        # logger.info("- 1 -----------------------------------------")

        # Build a complete World from Lite objects
        world_lite: World = (await self.worlddao.get(world_id=request.id)).data
        island_ids: set[str] = world_lite.ids
        # logger.info(f"island_ids={island_ids}")
        partial_world = world_lite.model_dump(exclude={"island_ids"})
        # logger.info("- 2 -----------------------------------------")
        # logger.info(partial_world)
        world: WorldFull = WorldFull.model_validate(partial_world)
        # logger.info("- 3 -----------------------------------------")
        # logger.info(world)
        # logger.info("- 4 -----------------------------------------")
        for island_id in island_ids:
            local_island: IslandFull = await self.island_get(
                request=IslandGetRequest(world_id=request.id, island_id=island_id)
            )
            world.contents[island_id] = local_island
        return world

    async def world_delete(self, request: WorldDeleteRequest) -> None:
        logger.debug("[StateService] deleting world")
        await self.worlddao.delete(world_id=request.id)

    ### Island ##################################

    async def island_create(self, request: IslandCreateRequest) -> str:
        logger.debug("[StateService] creating island")
        new_island: Island = await self.flatislandfactory.create(
            world_id=request.world_id, name=request.name, dimensions=request.dimensions, biome=request.biome
        )
        # add new island to world
        # add island to world and store

        # get
        wrapped_world_lite: WrappedData[World] = await self.worlddao.get(world_id=request.world_id)

        # patch
        wrapped_world_lite.data.ids.add(new_island.id)

        # put
        await self.worlddao.put_safe(wrapped_world=wrapped_world_lite)

        return new_island.id

    async def island_delete(self, request: IslandDeleteRequest) -> None:
        msg: str = f"[WorldService] deleting island {id}"
        logger.debug(msg)
        await self.islanddao.delete(world_id=request.world_id, island_id=request.island_id)

    async def island_lite_get(self, request: IslandGetRequest) -> Island:
        return (await self.islanddao.get(world_id=request.world_id, island_id=request.island_id)).data

    async def island_get(self, request: IslandGetRequest) -> IslandFull:
        # Builds a complete Island from Lite objects
        island_lite: Island = (await self.islanddao.get(world_id=request.world_id, island_id=request.island_id)).data
        tile_ids: set[str] = island_lite.ids
        island_partial = island_lite.model_dump(exclude={"tile_ids"})
        island: IslandFull = IslandFull.model_validate(island_partial)

        # re-hydrate the tiles
        for tile_id in tile_ids:
            tile: Tile = (await self.tiledao.get(world_id=request.world_id, island_id=island.id, tile_id=tile_id)).data
            island.contents[tile_id] = tile

        return island



from ..sdk.contracts.dtos.island import Island
from ..sdk.contracts.dtos.island_full import IslandFull
from ..sdk.contracts.dtos.sdk.command_options import CommandOptions
from ..sdk.contracts.dtos.sdk.requests.island.create import IslandCreateRequest
from ..sdk.contracts.dtos.sdk.requests.island.delete import IslandDeleteRequest
from ..sdk.contracts.dtos.sdk.requests.island.get import IslandGetRequest
from ..sdk.contracts.dtos.sdk.requests.world.create import WorldCreateRequest
from ..sdk.contracts.dtos.sdk.requests.world.delete import WorldDeleteRequest
from ..sdk.contracts.dtos.sdk.requests.world.get import WorldGetRequest
from ..sdk.contracts.dtos.sdk.responses.island.create import IslandCreateResponse
from ..sdk.contracts.dtos.sdk.responses.island.get import IslandGetResponse
from ..sdk.contracts.dtos.sdk.responses.world.create import WorldCreateResponse
from ..sdk.contracts.dtos.sdk.responses.world.get import WorldGetResponse
from ..sdk.contracts.dtos.world import World
from ..sdk.contracts.dtos.world_full import WorldFull
from ..sdk.contracts.types.tile import TileType
from .commands.island.create import IslandCreateCommand
from .commands.island.delete import IslandDeleteCommand
from .commands.island.get import IslandGetCommand
from .commands.metrics.health.get import HealthGetCommand
from .commands.world.create import WorldCreateCommand
from .commands.world.delete import WorldDeleteCommand
from .commands.world.get import WorldGetCommand


class Client:
    # metrics/health
    health_get_command: HealthGetCommand

    # island
    island_create_command: IslandCreateCommand
    island_delete_command: IslandDeleteCommand
    island_get_command: IslandGetCommand

    # world
    world_create_command: WorldCreateCommand
    world_get_command: WorldGetCommand
    world_delete_command: WorldDeleteCommand

    def __init__(self, options: CommandOptions | None = None):
        command_dict: dict = {}
        if options:
            command_dict["options"] = options

        ### metrics
        self.health_get_command = HealthGetCommand.model_validate(command_dict)

        ### island
        self.island_create_command = IslandCreateCommand.model_validate(command_dict)
        self.island_delete_command = IslandDeleteCommand.model_validate(command_dict)
        self.island_get_command = IslandGetCommand.model_validate(command_dict)

        # world
        self.world_create_command = WorldCreateCommand.model_validate(command_dict)
        self.world_get_command = WorldGetCommand.model_validate(command_dict)
        self.world_delete_command = WorldDeleteCommand.model_validate(command_dict)

    # metrics

    async def health_get(self) -> dict:
        """
        Gets the server health.

        Returns
        -------
        health: dict
            The server health (true or false).
        """

        return await self.health_get_command.execute()

    # island

    async def island_create(self, world_id: str, name: str | None, dimensions: tuple[int, int], biome: TileType) -> str:
        request: IslandCreateRequest = IslandCreateRequest(
            world_id=world_id, name=name, dimensions=dimensions, biome=biome
        )
        response: IslandCreateResponse = await self.island_create_command.execute(request=request)
        return response.id

    async def island_get(self, world_id: str, island_id: str, full: bool = False) -> IslandFull | Island:
        request: IslandGetRequest = IslandGetRequest(world_id=world_id, island_id=island_id, full=full)
        response: IslandGetResponse = await self.island_get_command.execute(request=request)
        return response.island

    async def island_delete(self, world_id: str, island_id: str) -> None:
        request: IslandDeleteRequest = IslandDeleteRequest(world_id=world_id, island_id=island_id)
        await self.island_delete_command.execute(request=request)

    # world

    async def world_create(self, name: str | None) -> str:
        request: WorldCreateRequest = WorldCreateRequest(name=name)
        response: WorldCreateResponse = await self.world_create_command.execute(request)
        return response.id

    async def world_delete(self, world_id: str) -> None:
        request: WorldDeleteRequest = WorldDeleteRequest(world_id=world_id)
        await self.world_delete_command.execute(request=request)

    async def world_get(self, world_id: str, full: bool) -> WorldFull | World:
        request: WorldGetRequest = WorldGetRequest(world_id=world_id, full=full)
        response: WorldGetResponse = await self.world_get_command.execute(request)
        return response.world



import logging
import traceback

from fastapi import APIRouter, HTTPException

from ....sdk.contracts.dtos.island import Island
from ....sdk.contracts.dtos.island_full import IslandFull
from ....sdk.contracts.dtos.sdk.requests.island.create import IslandCreateRequest
from ....sdk.contracts.dtos.sdk.requests.island.delete import IslandDeleteRequest
from ....sdk.contracts.dtos.sdk.requests.island.get import IslandGetRequest
from ....sdk.contracts.dtos.sdk.requests.world.create import WorldCreateRequest
from ....sdk.contracts.dtos.sdk.requests.world.delete import WorldDeleteRequest
from ....sdk.contracts.dtos.sdk.requests.world.get import WorldGetRequest
from ....sdk.contracts.dtos.sdk.responses.island.create import IslandCreateResponse
from ....sdk.contracts.dtos.sdk.responses.island.get import IslandGetResponse
from ....sdk.contracts.dtos.sdk.responses.response import Response
from ....sdk.contracts.dtos.sdk.responses.world.create import WorldCreateResponse
from ....sdk.contracts.dtos.sdk.responses.world.get import WorldGetResponse
from ....sdk.contracts.dtos.world import World
from ....sdk.contracts.dtos.world_full import WorldFull
from ....sdk.contracts.errors.server.dao.conflict import DaoConflictError
from ....sdk.contracts.errors.server.dao.doesnotexist import DaoDoesNotExistError
from ....sdk.contracts.errors.server.dao.inconsistency import DaoInconsistencyError
from ..context import ContextManager

# from pyinstrument import Profiler


logger = logging.getLogger()

router: APIRouter = APIRouter(
    prefix="/world",
    tags=["world"],
)


### /world


@router.get("/{world_id}")
async def world_get(world_id: str, full: bool = False) -> Response[WorldGetResponse]:
    request: WorldGetRequest = WorldGetRequest(id=world_id)

    try:
        if full:
            world: WorldFull = await ContextManager.state_service.world_get(request=request)
            response = Response[WorldGetResponse](data=WorldGetResponse(world=world))
        else:
            world_lite: World = await ContextManager.state_service.world_lite_get(request=request)
            response = Response[WorldGetResponse](data=WorldGetResponse(world=world_lite))
    except DaoConflictError as error:
        traceback.print_exc()
        logger.error(str(error))
        raise HTTPException(status_code=409, detail=str(error)) from error
    except DaoDoesNotExistError as error:
        traceback.print_exc()
        logger.error(str(error))
        raise HTTPException(status_code=404, detail=str(error)) from error
    except DaoInconsistencyError as error:
        traceback.print_exc()
        logger.error(str(error))
        raise HTTPException(status_code=500, detail=str(error)) from error
    except Exception as error:
        traceback.print_exc()
        logger.error(str(error))
        # catch everything else
        raise HTTPException(status_code=500, detail=f"Uncaught exception: {str(error)}") from error

    return response


@router.delete("/{world_id}")
async def world_delete(world_id: str) -> None:
    request: WorldDeleteRequest = WorldDeleteRequest(id=world_id)

    try:
        await ContextManager.state_service.world_delete(request=request)
    except DaoConflictError as error:
        traceback.print_exc()
        logger.error(str(error))
        raise HTTPException(status_code=409, detail=str(error)) from error
    except DaoDoesNotExistError as error:
        traceback.print_exc()
        logger.error(str(error))
        raise HTTPException(status_code=404, detail=str(error)) from error
    except DaoInconsistencyError as error:
        traceback.print_exc()
        logger.error(str(error))
        raise HTTPException(status_code=500, detail=str(error)) from error
    except Exception as error:
        traceback.print_exc()
        logger.error(str(error))
        # catch everything else
        raise HTTPException(status_code=500, detail=f"Uncaught exception: {str(error)}") from error


@router.post("")
async def world_create(request: WorldCreateRequest) -> Response[WorldCreateResponse]:
    try:
        world_id: str = await ContextManager.state_service.world_create(request=request)
    except DaoConflictError as error:
        traceback.print_exc()
        logger.error(str(error))
        raise HTTPException(status_code=409, detail=str(error)) from error
    except DaoDoesNotExistError as error:
        traceback.print_exc()
        logger.error(str(error))
        raise HTTPException(status_code=404, detail=str(error)) from error
    except DaoInconsistencyError as error:
        traceback.print_exc()
        logger.error(str(error))
        raise HTTPException(status_code=500, detail=str(error)) from error
    except Exception as error:
        traceback.print_exc()
        logger.error(str(error))
        # catch everything else
        raise HTTPException(status_code=500, detail=f"Uncaught exception: {str(error)}") from error

    response = Response[WorldCreateResponse](data=WorldCreateResponse(id=world_id))
    return response


### /world/world_id/island/


@router.post("/{world_id}/island")
async def island_create(world_id: str, island_create_request: IslandCreateRequest) -> Response[IslandCreateResponse]:
    # we ignore any passed in world_ids in this condition and over-write based on path.
    # Due to this world_id is optional within the DTO.
    island_create_request.world_id = world_id

    try:
        # with Profiler() as profiler:
        island_id: str = await ContextManager.state_service.island_create(request=island_create_request)
        # profiler.print()
    except DaoConflictError as error:
        traceback.print_exc()
        logger.error(str(error))
        raise HTTPException(status_code=409, detail=str(error)) from error
    except DaoDoesNotExistError as error:
        traceback.print_exc()
        logger.error(str(error))
        raise HTTPException(status_code=404, detail=str(error)) from error
    except DaoInconsistencyError as error:
        traceback.print_exc()
        logger.error(error)
        raise HTTPException(status_code=500, detail=str(error)) from error
    except Exception as error:
        traceback.print_exc()
        logger.error(str(error))
        # catch everything else
        raise HTTPException(status_code=500, detail=f"Uncaught exception: {str(error)}") from error

    return Response[IslandCreateResponse](data=IslandCreateResponse(id=island_id))


@router.get("/{world_id}/island/{island_id}")
async def island_get(world_id: str, island_id: str, full: bool = True) -> Response[IslandGetResponse]:
    request: IslandGetRequest = IslandGetRequest(world_id=world_id, island_id=island_id)

    try:
        if full:
            island: IslandFull = await ContextManager.state_service.island_get(request=request)
            response = Response[IslandGetResponse](data=IslandGetResponse(island=island))
        else:
            island_lite: Island = await ContextManager.state_service.island_lite_get(request=request)
            response = Response[IslandGetResponse](data=IslandGetResponse(island=island_lite))
    except DaoConflictError as error:
        traceback.print_exc()
        logger.error(str(error))
        raise HTTPException(status_code=409, detail=str(error)) from error
    except DaoDoesNotExistError as error:
        traceback.print_exc()
        logger.error(str(error))
        raise HTTPException(status_code=404, detail=str(error)) from error
    except DaoInconsistencyError as error:
        traceback.print_exc()
        logger.error(str(error))
        raise HTTPException(status_code=500, detail=str(error)) from error
    except Exception as error:
        traceback.print_exc()
        logger.error(str(error))
        traceback.print_exc()
        raise HTTPException(status_code=500, detail=f"Uncaught exception: {str(error)}") from error

    return response


@router.delete("/{world_id}/island/{island_id}")
async def island_delete(world_id: str, island_id: str) -> None:
    request: IslandDeleteRequest = IslandDeleteRequest(world_id=world_id, island_id=island_id)
    try:
        await ContextManager.state_service.island_delete(request=request)
    except DaoConflictError as error:
        traceback.print_exc()
        logger.error(str(error))
        raise HTTPException(status_code=409, detail=str(error)) from error
    except DaoDoesNotExistError as error:
        traceback.print_exc()
        logger.error(str(error))
        raise HTTPException(status_code=404, detail=str(error)) from error
    except DaoInconsistencyError as error:
        traceback.print_exc()
        logger.error(str(error))
        raise HTTPException(status_code=500, detail=str(error)) from error
    except Exception as error:
        traceback.print_exc()
        logger.error(str(error))
        # catch everything else
        raise HTTPException(status_code=500, detail=f"Uncaught exception: {str(error)}") from error



import logging
import os
import uuid
from pathlib import Path

from pydantic import BaseModel

from ...sdk.contracts.dtos.sdk.wrapped_data import WrappedData
from ...sdk.contracts.dtos.tiles.tile import Tile
from ...sdk.contracts.errors.server.dao.conflict import DaoConflictError
from ...sdk.contracts.errors.server.dao.doesnotexist import DaoDoesNotExistError
from ...sdk.contracts.errors.server.dao.inconsistency import DaoInconsistencyError

logger = logging.getLogger()


class TileDao(BaseModel):
    # ["base"] / "worlds" / "wold_id" / "islands" / "island_id" / "tiles" / "tile_id.json"
    storage_base_path: Path

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.storage_base_path.mkdir(parents=True, exist_ok=True)

    ### Internal ##################################

    def _tile_path(self, world_id: str, island_id: str, tile_id: str) -> Path:
        # ["base"] / "worlds" / "wold_id" / "islands" / "island_id" / "tiles" / "tile_id.json"
        return self.storage_base_path / "worlds" / world_id / "islands" / island_id / "tiles" / f"{tile_id}.json"

    async def get(self, world_id: str, island_id: str, tile_id: str) -> WrappedData[Tile]:
        logger.debug("[TileDAO] getting tile data from storage")
        tile_metadata_path: Path = self._tile_path(world_id=world_id, island_id=island_id, tile_id=tile_id)
        if not tile_metadata_path.exists():
            raise DaoDoesNotExistError("tile metadata does not exist")
        with open(file=tile_metadata_path, mode="r", encoding="utf-8") as file:
            os.fsync(file)
            json_data: str = file.read()
        return WrappedData[Tile].model_validate_json(json_data)

    async def post(self, world_id: str, island_id: str, tile: Tile) -> None:
        logger.debug("[TileDAO] posting tile data to storage")
        tile_metadata_path: Path = self._tile_path(world_id=world_id, island_id=island_id, tile_id=tile.id)
        if tile_metadata_path.exists():
            raise DaoConflictError("tile metadata already exists")
        if not tile_metadata_path.parents[1].exists():
            raise DaoDoesNotExistError("tile container (island) does not exist")
        if not tile_metadata_path.parent.exists():
            logger.debug("[TileDAO] tile metadata folder creating ..")
            tile_metadata_path.parent.mkdir(parents=True, exist_ok=True)
        nonce: str = str(uuid.uuid4())
        wrapped_data: WrappedData[Tile] = WrappedData[Tile](data=tile, nonce=nonce)
        wrapped_data_raw: str = wrapped_data.model_dump_json(exclude={"data": {"contents"}}, exclude_none=True)
        with open(file=tile_metadata_path, mode="w", encoding="utf-8") as file:
            file.write(wrapped_data_raw)
            os.fsync(file)

            # now validate we stored
        stored_tile: WrappedData[Tile] = await self.get(world_id=world_id, island_id=island_id, tile_id=tile.id)
        if stored_tile.nonce != nonce:
            msg: str = f"storage inconsistency detected while storing tile {tile.id} - nonce mismatch!"
            raise DaoInconsistencyError(msg)

    async def put_safe(self, world_id: str, island_id: str, wrapped_tile: WrappedData[Tile]) -> None:
        logger.debug("[TileDAO] putting tile data to storage")
        tile_metadata_path: Path = self._tile_path(world_id=world_id, island_id=island_id, tile_id=wrapped_tile.data.id)
        if not tile_metadata_path.parents[1].exists():
            raise DaoDoesNotExistError("tile container (island) does not exist")
        if not tile_metadata_path.parent.exists():
            logger.debug("[TileDAO] tile metadata folder creating ..")
            tile_metadata_path.parent.mkdir(parents=True, exist_ok=True)

        # see if we have a pre-existing nonce to verify against
        try:
            previous_state: WrappedData[Tile] = await self.get(
                world_id=world_id, island_id=island_id, tile_id=wrapped_tile.data.id
            )
            if previous_state.nonce != wrapped_tile.nonce:
                msg: str = f"storage inconsistency detected while putting tile {wrapped_tile.data.id} - nonce mismatch!"
                raise DaoInconsistencyError(msg)
        except DaoDoesNotExistError:
            # then no nonce to verify against.
            pass

        # if we made it this far we are safe to update

        nonce: str = str(uuid.uuid4())
        wrapped_data: WrappedData[Tile] = WrappedData[Tile](data=wrapped_tile.data, nonce=nonce)
        wrapped_data_raw: str = wrapped_data.model_dump_json(exclude={"data": {"contents"}}, exclude_none=True)
        with open(file=tile_metadata_path, mode="w", encoding="utf-8") as file:
            file.write(wrapped_data_raw)
            os.fsync(file)

        # now validate we stored
        stored_tile: WrappedData[Tile] = await self.get(
            world_id=world_id, island_id=island_id, tile_id=wrapped_data.data.id
        )
        if stored_tile.nonce != nonce:
            msg: str = (
                f"storage inconsistency detected while verifying put tile {wrapped_data.data.id} - nonce mismatch!"
            )
            raise DaoInconsistencyError(msg)

    async def delete(self, world_id: str, island_id: str, tile_id: str) -> None:
        logger.debug("[TileDAO] deleting tile data from storage")
        tile_metadata_path: Path = self._tile_path(world_id=world_id, island_id=island_id, tile_id=tile_id)
        if not tile_metadata_path.exists():
            raise DaoDoesNotExistError("tile metadata does not exist")
        os.remove(path=tile_metadata_path)
